#include <sta/devices/raspi/bus/i2c.hpp>

#ifdef STA_PLATFORM_RASPI

#include <sta/debug/assert.hpp>
#include <sta/debug/debug.hpp>

#include <cstdlib>
#include <string.h>
#include <iostream>

#include <linux/i2c-dev.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>


namespace sta 
{
    RaspiI2C::RaspiI2C(I2CNode node, Mutex * mutex, bool persistent_open)
        : I2C{mutex}, persistent_open_{persistent_open}
    {
        // Safer version of malloc + strcpy
        i2cdev_ = strdup(node == I2CNode::DEV_1 ? "/dev/i2c-1" : "/dev/i2c-2");

        STA_ASSERT(i2cdev_ != nullptr);
    }

    RaspiI2C::~RaspiI2C()
    {
        if (i2cdev_ != NULL ) {
            free(i2cdev_);
	        i2cdev_ = NULL;
        }
        
        if (open_) {
            close(i2cfd_);
        }
    }

    void RaspiI2C::transfer(uint8_t value)
    {
        STA_ASSERT(open_);

        selectAddress();
        ssize_t n_out = write(i2cfd_, &value, 1);

        STA_ASSERT(n_out == 1);
    }

    void RaspiI2C::transfer16(uint16_t value)
    {
        STA_ASSERT(open_);

        selectAddress();
        ssize_t n_out = write(i2cfd_, &value, 2);

        STA_ASSERT(n_out == 2);
    }

    void RaspiI2C::transfer(const uint8_t * buffer, size_t size)
    {
        STA_ASSERT(open_);

        selectAddress();
        ssize_t n_out = write(i2cfd_, buffer, size);

        STA_ASSERT(n_out == size);
    }

    void RaspiI2C::receive(uint8_t * buffer, size_t size)
    {
        STA_ASSERT(open_);

        selectAddress();
        ssize_t n_in = read(i2cfd_, buffer, size);

        STA_ASSERT(n_in >= 0);
    }

    void RaspiI2C::fill(uint8_t value, size_t count) 
    {
        STA_ASSERT(open_);

        selectAddress();

        // Initialize a buffer of size count and fill it with the value.
        uint8_t *buffer = new uint8_t[count];
        memset(buffer, value, count);

        ssize_t n_out = write(i2cfd_, buffer, count);

        STA_ASSERT(n_out == count);

        delete [] buffer; 
    }

    void RaspiI2C::selectAddress()
    {
        // Select the slave with the given address.
        int rslt = ioctl(i2cfd_, I2C_SLAVE, address_);

        STA_ASSERT(rslt != -1);
    }

    void RaspiI2C::acquire()
    {
        I2C::acquire();

        if (open_) {
            return;
        }

        i2cfd_ = open(i2cdev_, O_RDWR);
        open_ = true;

        STA_ASSERT(i2cfd_ >= 0);
    }

    void RaspiI2C::release()
    {
        if (!persistent_open_ && open_) {
            close(i2cfd_);
            open_ = false;
        }

        I2C::release();
    }

    RaspiI2CDevice::RaspiI2CDevice(RaspiI2C * intf, uint16_t address_10bit, bool master, bool blocking) 
        : I2CDevice { intf, address_10bit, master, blocking }
    {
        
    }


} // namespace sta

#endif // STA_PLATFORM_RASPI
